# __version__.py

# autogenerated by keats 0.2.29

__version__ = "1.0.9"
__name__ = "pydent"
__title__ = "pydent"
__authors__ = ['jvrana <justin.vrana@gmail.com>', 'Ben Keller <bjkeller@uw.edu>', 'Eric Klavins <klavins@uw.edu>']
__repository__ = "https://www.github.com/aquariumbio/trident"
__homepage__ = "https://www.github.com/aquariumbio/trident"
__description__ = "Aquarium's Python API for planning, executing, and analyzing scientific experiments."
__license__ = "MIT"
__maintainers__ = None
__readme__ = "README.md"